# frozen_string_literal: true

require 'import'
require 'cht-lib-fabric/crn'
require 'securerandom'
require 'flow_down/errors'


module FlowDown
  module Data
    class PersistentStore
      include Import['logger',
                     'dynamodb.client',
                     'dynamodb.flow_down_table',
                     'flow_down.data.expression_helper']

      # Where the flow down data is persistently stored
      # 400KB Size Limit per item
      # https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Limits.html
      # Parition Key Info:
      # https://aws.amazon.com/blogs/database/choosing-the-right-dynamodb-partition-key/
      # Global Secondary Indexes:
      # https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/GSI.html
      # https://aws.amazon.com/blogs/database/how-to-use-dynamodb-global-secondary-indexes-to-improve-query-performance-and-reduce-costs/
      #
      # TABLE SCHEMA (Easy / more flexible to query, fits more easily in 400KB per item boundary)
      # {
      #    content_id: UUID
      #    id: S
      #    customer_id: S
      #    organization_id: S
      #    content_type: S
      #    enabled: B
      # }
      #
      # Get a set of items enabled for flow down from persistent storage
      #
      # Use Cases Supported
      # - Query for all flow down by Customer and Organization
      # - Query for all flow down of a content type by Customer and Organization
      # - Query for flow down of an ID by Customer and Organization
      #
      def get_flow_down_items(input, customer_id, organization_id)
        logger.info "Fetching flow down items from persistent store table [#{flow_down_table.name}]"
        # Query for items (Using GSI - Global Secondary Index)
        # because our main queries are looking for groups of items
        query = expression_helper.build_base_query(input, customer_id, organization_id)
        flow_down_items = client.query(query)
        if flow_down_items.nil?
          logger.debug "No flow down items found for customer [#{customer_id}] org [#{organization_id}]"
          []
        else
          logger.debug "Found flow down items #{flow_down_items[:items]} for customer [#{customer_id}] org [#{organization_id}]"
          flow_down_items[:items]
        end
      end

      #
      # Enable Flow Down Persistently for a set of content
      #
      def enable_flow_down(input, customer_id, organization_id, user_id)
        logger.info "Persisting flow down items into persistent store table [#{flow_down_table.name}]"
        items_changed = []
        items = input.each do |item|
          logger.debug "Item #{item.to_json}"
          crn = ::Platform::CRN.from_s(item[:id])
          content_type = crn.service
          if item[:enabled]
            logger.debug "Enabling flow down item #{item.to_json} of type #{content_type}"
            # Check to see if its already enabled to avoid duplicates
            raise ::FlowDown::Exceptions::AlreadyExists, "FlowDown item already exists" if !item_exists(item, customer_id, organization_id).nil?
            # Add an item when its enabled
            new_item = {
              id: item[:id],
              content_id: SecureRandom.uuid,
              customer_id: customer_id.to_s,
              organization_id: organization_id.to_s,
              content_type: content_type,
              enabled: item[:enabled],
              enabled_by_id: user_id.to_s,
              enabled_at: DateTime.now.iso8601
            }
            client.put_item(
              item: new_item,
              table_name: flow_down_table.name
            )
            items_changed << new_item
          else
            logger.debug "Disabling flow down item #{item.to_json} of type #{content_type}"
            # Since we have to delete by primary key, we have to find the record first
            item_to_delete = item_exists(item, customer_id, organization_id)
            raise ::FlowDown::Exceptions::NotFound, "FlowDown item not found" if item_to_delete.nil?
            logger.debug "Items to disable flow down for #{item.to_json}"
            # Remove item when its disabled
            client.delete_item(
              key: {
                content_id: item_to_delete['content_id'],
                organization_id: item_to_delete['organization_id'].to_s
              },
              table_name: flow_down_table.name
            ) unless item_to_delete.nil?
            items_changed << item_to_delete unless item_to_delete.nil?
          end
        end
        items_changed
      end

      #
      # Returns an a boolean indicating whether an item exists already
      #
      def item_exists(item, customer_id, organization_id)
        item = client.query(
          table_name: flow_down_table.name,
          index_name: 'flowdown_items',
          key_condition_expression: 'customer_id = :customer_id and organization_id = :organization_id',
          expression_attribute_values: {
            ':customer_id' => customer_id.to_s,
            ':organization_id' => organization_id.to_s,
            ':id' => item[:id]
          },
          filter_expression: 'id = :id'
        )
        return item[:items][0]
      end
    end
  end
end
