# frozen_string_literal: true
require 'opentracing'

module FlowDown
  # Mixin to wrap a method with tracing.
  #
  # @example
  #   class SomeClass
  #     extend Traceable
  #     def some_method ... end
  #     trace_method :some_method
  #  end
  module Traceable
    # Wrap a method with OpenTracing tracing.
    #
    # For each named method, replace that method with a wrapper that
    # creates an OpenTracing span describing a call to it.  The span
    # is named after the class and method name, and with no
    # non-default tags.
    def trace_method(*names)
      names.each do |name|
        meth = instance_method(name)
        define_method(name) do |*args|
          cls = self.class.name.split('::').last
          scope = ::OpenTracing.start_active_span("#{cls}.#{__callee__}")
          begin
            meth.bind(self).call(*args)
          ensure
            scope.close
          end
        end
      end
    end
  end
end